"""
PyDetex
https://github.com/ppizarror/PyDetex

GUI
Basic gui that convers and executes a given pipeline.
"""

__all__ = ['PyDetexGUI']

import concurrent.futures
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox

import pyperclip
import requests
import string
import sys

from nltk.tokenize import RegexpTokenizer
from outdated import check_outdated
from PyMultiDictionary import MultiDictionary
from typing import Optional, Tuple
from warnings import warn

sys.path.append('../')

import pydetex._gui_utils as gui_ut
import pydetex.pipelines as pip
import pydetex.utils as ut
import pydetex.version

from pydetex._fonts import FONT_TAGS
from pydetex._gui_settings import Settings
from pydetex.parsers import FONT_FORMAT_SETTINGS as PARSER_FONT_FORMAT

# Settings
_MAX_PASTE_RETRY: int = 3


class PyDetexGUI(object):
    """
    GUI.
    """

    _cfg: 'Settings'
    _copy_clip: 'tk.Button'
    _detect_language_event_id: str
    _detected_lang_tag: str
    _dictionary: 'MultiDictionary'
    _dictionary_window: Optional['gui_ut.DictionaryGUI']
    _dictionary_btn: 'tk.Button'
    _paste_timeout_error: int
    _ready: bool
    _root: 'tk.Tk'
    _settings_window: Optional['gui_ut.SettingsWindow']
    _status_bar_cursor: 'tk.Label'
    _status_bar_cursor_sel: 'tk.Label'
    _status_bar_lang: 'tk.Label'
    _status_bar_status: 'tk.Label'
    _status_bar_words: 'tk.Label'
    _status_clear_event_id: str
    _text_in: 'gui_ut.RichText'
    _text_out: 'gui_ut.RichText'
    _tokenizer: 'RegexpTokenizer'

    def __init__(self) -> None:
        """
        Constructor.
        """

        # ----------------------------------------------------------------------
        # Creates the window
        # ----------------------------------------------------------------------
        self._root = tk.Tk()

        # Dictionary
        self._dictionary = MultiDictionary()
        self._dictionary_window = None

        # Load settings
        self._cfg = Settings()
        window_size = self._cfg.get(self._cfg.CFG_WINDOW_SIZE).copy()

        # Configure window
        self._root.title('PyDetex')
        img = tk.Image('photo', file=ut.RESOURCES_PATH + 'icon.gif')
        # noinspection PyProtectedMember
        self._root.tk.call('wm', 'iconphoto', self._root._w, img)
        if not ut.IS_OSX:
            self._root.iconbitmap(ut.RESOURCES_PATH + 'icon.ico')

        self._root.minsize(width=window_size[0], height=window_size[1])
        self._root.resizable(width=False, height=False)
        gui_ut.center_window(self._root, window_size)
        self._root.protocol('WM_DELETE_WINDOW', self._close)

        # ----------------------------------------------------------------------
        # Settings button
        # ----------------------------------------------------------------------
        f0 = tk.Frame(self._root, border=10, width=window_size[0], height=50)
        f0.pack()
        f0.pack_propagate(0)
        tk.Button(f0, text=ut.button_text(self._cfg.lang('about')), command=self._about,
                  relief=tk.GROOVE).pack(side=tk.RIGHT)
        tk.Button(f0, text=ut.button_text(self._cfg.lang('settings')), command=self._open_settings,
                  relief=tk.GROOVE).pack(side=tk.RIGHT, padx=(0, 7 if ut.IS_OSX else 10))
        self._dictionary_btn = tk.Button(f0, text=ut.button_text(self._cfg.lang('dictionary')),
                                         command=self._open_dictionary, relief=tk.GROOVE)
        self._dictionary_btn.pack(side=tk.RIGHT, padx=(0, 7 if ut.IS_OSX else 10))

        hthick, hcolor = 3 if ut.IS_OSX else 1, '#426392' if ut.IS_OSX else '#475aff'
        fsize = self._cfg.get(self._cfg.CFG_FONT_SIZE)

        # ----------------------------------------------------------------------
        # Input texts
        # ----------------------------------------------------------------------
        # In text
        f1 = tk.Frame(self._root, border=0, width=window_size[0], height=window_size[2])
        f1.pack(fill='both', padx=10)
        f1.pack_propagate(0)

        self._text_in = gui_ut.RichText(self._cfg, f1, wrap='word', highlightthickness=hthick,
                                        highlightcolor=hcolor, font_size=fsize, editable=True)
        self._text_in.pack(fill='both')
        self._text_in.bind('<Button>', self._process_cursor_in)
        self._text_in.bind('<ButtonRelease>', self._process_cursor_in)
        self._text_in.bind('<FocusIn>', self._process_focusin_in)
        self._text_in.bind('<FocusOut>', self._process_focusout_in)
        self._text_in.bind('<Key>', self._process_in_key)

        # Out text
        f2 = tk.Frame(self._root, border=0, width=window_size[0], height=window_size[2])
        f2.pack(fill='both', padx=10, pady=5)
        f2.pack_propagate(0)

        self._text_out = gui_ut.RichText(self._cfg, f2, wrap='word', highlightthickness=hthick,
                                         highlightcolor=hcolor, font_size=fsize, copy=True)
        self._text_out.bind('<Key>', self._process_out_key)
        self._text_out.bind('<Button>', self._check_dictionary_selected)
        self._text_out.bind('<ButtonRelease>', self._check_dictionary_selected)
        self._text_out.bind('<FocusIn>', self._check_dictionary_selected)
        self._text_out.bind('<FocusOut>', self._check_dictionary_selected)
        self._text_out.pack(fill='both')

        # ----------------------------------------------------------------------
        # Commands buttons
        # ----------------------------------------------------------------------
        command_btn_packx = 15  # margin px
        f3 = tk.Frame(self._root, border=2)
        f3.pack(pady=(9, 18))

        # Process
        ut.Button(f3, text=ut.button_text(self._cfg.lang('process')), command=self._process, relief=tk.GROOVE,
                  bg='#475aff' if ut.IS_OSX else '#6388ff').pack(side=tk.LEFT, padx=(0, command_btn_packx))

        # Process clip
        tk.Button(f3, text=ut.button_text(self._cfg.lang('process_clip')), command=self._process_clip,
                  relief=tk.GROOVE).pack(side=tk.LEFT, padx=(0, command_btn_packx))

        # Copy to clip
        self._copy_clip = tk.Button(f3, text=ut.button_text(self._cfg.lang('process_copy')), command=self._copy_to_clip,
                                    relief=tk.GROOVE)
        self._copy_clip.pack(side=tk.LEFT, padx=(0, command_btn_packx))

        # Clear
        ut.Button(f3, text=ut.button_text(self._cfg.lang('clear')), command=self._clear,
                  relief=tk.GROOVE, bg='#ff7878').pack(side=tk.LEFT)

        # ----------------------------------------------------------------------
        # Status bar
        # ----------------------------------------------------------------------
        ttk.Separator(self._root, orient='horizontal').pack(side='top', fill='x')
        status_fg = '#999999' if ut.IS_OSX else '#666666'

        f4 = tk.Frame(self._root, width=window_size[0], height=26)
        f4.pack(fill='both')
        f4.pack_propagate(0)

        # Detected language
        show_status = 0.2 if window_size[0] > 750 else 0
        self._status_bar_lang = gui_ut.make_label(f4, w=window_size[0] * (0.4 + (0.2 - show_status)), h=20,
                                                  side=tk.LEFT, fg=status_fg, bd=0, relief=tk.SUNKEN, anchor=tk.W,
                                                  pad=(0, 0, 0, 10), separator=True)

        # Status
        self._status_bar_status = gui_ut.make_label(f4, w=window_size[0] * show_status, h=20, side=tk.LEFT,
                                                    fg=status_fg, bd=0, relief=tk.SUNKEN, anchor=tk.W, pad=(0, 0, 0, 5),
                                                    separator=True)

        # Cursor
        self._status_bar_cursor = gui_ut.make_label(f4, w=window_size[0] * 0.11, h=20, side=tk.LEFT, fg=status_fg,
                                                    bd=0, relief=tk.SUNKEN, anchor=tk.W, pad=(0, 0, 0, 5))

        # Cursor selected
        self._status_bar_cursor_sel = gui_ut.make_label(f4, w=window_size[0] * 0.14, h=20, side=tk.LEFT, fg=status_fg,
                                                        bd=0, relief=tk.SUNKEN, anchor=tk.W, pad=(0, 0, 0, 5),
                                                        separator=True)

        # Total processed words
        self._status_bar_words = gui_ut.make_label(f4, w=window_size[0] * 0.15, h=20, side=tk.LEFT, fg=status_fg,
                                                   bd=0, relief=tk.SUNKEN, anchor=tk.W, pad=(0, 0, 0, 5))

        # ----------------------------------------------------------------------
        # Final settings
        # ----------------------------------------------------------------------
        self._status_clear_event_id = ''
        self._clear()  # This also changes states

        # Set variables
        self._detect_language_event_id = ''
        self._detected_lang_tag = '–'
        self._paste_timeout_error = 0
        self._ready = False
        self._settings_window = None
        self._tokenizer = RegexpTokenizer(r'\w+')

        # Inserts the placeholder text
        self._insert_in(self._cfg.lang('placeholder'))
        self._root.update()

    def _get_selected_word(self) -> str:
        """
        Get selected word.

        :return: Return the current selected word
        """
        try:
            if self._text_in.selection_get() != '' and self._root.focus_get() == self._text_in:
                sf = self._text_in.count('1.0', 'sel.first')
                if sf is None:
                    sf = (0,)
                word = ut.get_word_from_cursor(self._text_in.get(0.0, tk.END), sf[0])
                if word != '':
                    return word
        except tk.TclError:
            pass
        try:
            if self._text_out.selection_get() != '' and self._root.focus_get() == self._text_out:
                sf = self._text_out.count('1.0', 'sel.first')
                if sf is None:
                    sf = (0,)
                word = ut.get_word_from_cursor(self._text_out.get(0.0, tk.END), sf[0])
                if word != '':
                    return word
        except tk.TclError:
            pass
        return ''

    # noinspection PyUnusedLocal
    def _check_dictionary_selected(self, *args) -> str:
        """
        Check dictionary selected word.

        :return: Return the selected word
        """
        word = ut.tokenize(self._get_selected_word())
        return word

    def _insert_in(self, text: str) -> None:
        """
        Insert text to in widget.

        :param text: Text
        """
        self._text_in.insert(tk.END, text)
        self._detect_language()
        self._process_cursor_in(None)

    def _status(self, text: str, clear: bool = False, clear_time: int = 500) -> None:
        """
        Set the app status.

        :param text: Status text
        :param clear: Clear after time
        :param clear_time: Clear time in miliseconds
        """
        if self._status_clear_event_id != '':
            self._root.after_cancel(self._status_clear_event_id)
        self._status_bar_status['text'] = text
        if clear:
            self._status_clear_event_id = self._root.after(clear_time, self._status_clear)

    def _status_clear(self) -> None:
        """
        Clear the status text.
        """
        self._status(self._cfg.lang('status_idle'))

    # noinspection PyUnresolvedReferences
    def _process_in_key(self, event: 'tk.Event') -> Optional['tk.Event']:
        """
        Process in keys.

        :param event: Event
        :return: Event
        """
        if self._detect_language_event_id != '':
            self._root.after_cancel(self._detect_language_event_id)

        if event.char in string.printable and event.char != '':
            self._status(self._cfg.lang('status_writing'), True)
        else:
            self._status_clear()
        self._detect_language_event_id = self._root.after(100, self._detect_language)
        self._process_cursor_in(event)
        return event

    def _process_cursor_in(self, event: Optional['tk.Event']) -> Optional['tk.Event']:
        """
        Process cursor on input text.

        :param event: Event
        :return: Event
        """
        self._root.after(50, self._process_cursor_event)
        return event

    def _process_focusin_in(self, event: Optional['tk.Event']) -> Optional['tk.Event']:
        """
        Process focus in on input text.

        :param event: Event
        :return: Event
        """
        self._process_cursor_in(event)
        return event

    def _process_focusout_in(self, event: Optional['tk.Event']) -> Optional['tk.Event']:
        """
        Process focus out on input text.

        :param event: Event
        :return: Event
        """
        w = self._cfg.get(self._cfg.CFG_WINDOW_SIZE)[0] > 750
        fout = self._cfg.lang('status_cursor_input_focusout') if w else self._cfg.lang(
            'status_cursor_input_focusout_min')
        self._status_bar_cursor['text'] = fout
        self._status_bar_cursor_sel['text'] = ''
        return event

    def _process_cursor_event(self) -> None:
        """
        Process cursor position.
        """
        window_w = self._cfg.get(self._cfg.CFG_WINDOW_SIZE)[0]
        cur_pos = self._text_in.index(tk.INSERT)

        # Check cursor position
        if cur_pos is None:
            self._status_bar_cursor['text'] = self._cfg.lang('status_cursor_null')
        else:
            line, pos = tuple(cur_pos.split('.'))
            cur_t = self._cfg.lang('status_cursor') if window_w > 750 else self._cfg.lang('status_cursor_min')
            self._status_bar_cursor['text'] = cur_t.format(pos, line)

        # Get cursor selection
        try:
            sf = self._text_in.count('1.0', 'sel.first')
            sl = self._text_in.count('1.0', 'sel.last')
            if sf is None:
                sf = (0,)
            d = sl[0] - sf[0]
            sel = self._cfg.lang('status_cursor_selected') if window_w > 750 else self._cfg.lang(
                'status_cursor_selected_min')
            chars = self._cfg.lang('status_cursor_selected_chars') if window_w > 900 else self._cfg.lang(
                'status_cursor_selected_chars_min')
            s = f'{sel}: '
            text = self._text_in.get(0.0, tk.END)
            if d == 1:
                self._status_bar_cursor_sel['text'] = \
                    s + self._cfg.lang('status_cursor_selected_chars_single')
            elif d == len(text) or d == len(text.strip()):
                self._status_bar_cursor_sel['text'] = self._cfg.lang('status_cursor_selected_all')
            else:
                self._status_bar_cursor_sel['text'] = s + chars.format(d)

        except tk.TclError:
            self._status_bar_cursor_sel['text'] = ''

        self._check_dictionary_selected()

    def _process_out_key(self, event: 'tk.Event') -> Optional['tk.Event']:
        """
        Process out keys.

        :param event: Event
        :return: Event
        """
        self._check_dictionary_selected()
        # noinspection PyUnresolvedReferences
        if event.char == '':
            return event

    def _detect_language(self) -> None:
        """
        Detects the input lang.
        """
        text = self._text_in.get(0.0, tk.END).strip()

        # Check language
        self._detected_lang_tag = ut.detect_language(text)
        lang = self._dictionary.get_language_name(self._detected_lang_tag, self._cfg.get(self._cfg.CFG_LANG))
        if text != '':
            self._status_bar_lang['text'] = self._cfg.lang('detected_lang').format(lang, self._detected_lang_tag)
        else:
            self._status_bar_lang['text'] = self._cfg.lang('detected_lang_write')
        self._detect_language_event_id = ''

        # Check if dictionary is available
        # noinspection PyProtectedMember
        if self._detected_lang_tag in self._dictionary._langs.keys():
            self._status_bar_lang['text'] += f'  [{self._cfg.lang("dictionary")}]'
            self._dictionary_btn['state'] = tk.NORMAL
        else:
            self._dictionary_btn['state'] = tk.DISABLED

    def start(self) -> None:
        """
        Starts the application.
        """
        # noinspection PyProtectedMember
        if self._cfg._last_opened_day_diff >= 7:
            self._root.after(1000, self._check_version_event)
        self._root.mainloop()

    def _clear(self) -> None:
        """
        Clear texts.
        """
        self._text_out['state'] = tk.NORMAL
        self._text_in.delete(0.0, tk.END)
        self._text_out.delete(0.0, tk.END)
        self._text_out['state'] = tk.DISABLED
        self._copy_clip['state'] = tk.DISABLED
        self._ready = False
        self._detect_language()
        self._status_bar_words['text'] = self._cfg.lang('status_words').format(0)
        self._status_clear()
        self._process_cursor_event()
        self._text_in.focus_force()

    @property
    def pipeline(self) -> pip.PipelineType:
        """
        Return the pipeline.

        :return: Pipeline
        """
        return self._cfg.get(self._cfg.CFG_PIPELINE)

    def _process(self) -> None:
        """
        Process and call the pipeline.
        """
        text = self._text_in.get(0.0, tk.END)
        if text.strip() == '':
            return self._clear()
        self._status(self._cfg.lang('status_processing'), True)

        self._text_out['state'] = tk.NORMAL
        self._copy_clip['state'] = tk.NORMAL

        # Font format
        font_format = self._cfg.get(self._cfg.CFG_OUTPUT_FONT_FORMAT)
        PARSER_FONT_FORMAT['cite'] = FONT_TAGS['link'] if font_format else ''
        PARSER_FONT_FORMAT['normal'] = FONT_TAGS['normal'] if font_format else ''
        PARSER_FONT_FORMAT['ref'] = FONT_TAGS['link'] if font_format else ''

        # Process the text and get the language
        out = self.pipeline(text)
        words = len(self._tokenizer.tokenize(out))
        self._cfg.add_words(words)

        # Check repeated words
        if self._cfg.get(self._cfg.CFG_CHECK_REPETITION):
            out = ut.check_repeated_words(
                s=out,
                lang=self._detected_lang_tag,
                min_chars=self._cfg.get(self._cfg.CFG_REPETITION_MIN_CHAR),
                window=self._cfg.get(self._cfg.CFG_REPETITION_DISTANCE),
                stopwords=self._cfg.get(self._cfg.CFG_REPETITION_USE_STOPWORDS),
                stemming=self._cfg.get(self._cfg.CFG_REPETITION_USE_STEMMING),
                ignore=self._tokenizer.tokenize(self._cfg.get(self._cfg.CFG_REPETITION_IGNORE_WORDS)),
                font_tag_format=FONT_TAGS['repeated_tag'] if font_format else '',
                font_param_format=FONT_TAGS['repeated_word'] if font_format else '',
                font_normal_format=FONT_TAGS['normal'] if font_format else '',
                remove_tokens=list(FONT_TAGS.values()),
                tag=self._cfg.lang('tag_repeated')
            )

        # Apply syntax highlight
        out = ut.syntax_highlight(out)

        # Insert the text
        self._text_out.insert_highlighted_text(out, True, font_format)

        # Configure status
        self._status_bar_words['text'] = self._cfg.lang('status_words').format(words)
        self._ready = True
        self._text_out['state'] = tk.DISABLED
        self._detect_language()

    def _process_clip(self) -> None:
        """
        Process from clipboard. Tries with pooling.
        """

        def _paste():
            return pyperclip.paste()

        self._status(self._cfg.lang('copy_from_clip'))
        executor = concurrent.futures.ThreadPoolExecutor(1)
        future = executor.submit(_paste)
        try:
            text = future.result(timeout=1)
        except concurrent.futures.TimeoutError:
            self._paste_timeout_error += 1
            if self._paste_timeout_error <= _MAX_PASTE_RETRY:
                # print(f'Paste process failed (TimeoutError), retrying {self._paste_timeout_error}/{_MAX_PASTE_RETRY}')
                self._root.after(100, self._process_clip)
            else:
                self._paste_timeout_error = 0
                error = f'Paste process failed after {_MAX_PASTE_RETRY} attempts'
                warn(error)
            return

        self._paste_timeout_error = 0
        if text.strip() == '':
            return
        self._text_in.delete(0.0, tk.END)
        self._text_in.insert(0.0, text)
        self._process()

    def _get_pipeline_results(self) -> str:
        """
        Returns the pipeline results.

        :return: Text
        """
        return str(self._text_out.get(0.0, tk.END)).strip()

    def _copy_to_clip(self) -> None:
        """
        Copy results to clip.
        """
        text = self._text_out.get(0.0, tk.END)
        pyperclip.copy(text)

    def _open_settings(self) -> None:
        """
        Launch settings.
        """
        if self._settings_window:
            self._settings_window.root.lift()
            return
        self._settings_window = gui_ut.SettingsWindow((375, 427 if ut.IS_OSX else 448), self._cfg)
        self._settings_window.on_destroy = self._close_settings
        try:
            # self._settings_window.root.mainloop(1)
            self._settings_window.root.update()
        except AttributeError:
            pass

    def _close_settings(self) -> None:
        """
        Close settings.
        """
        self._settings_window = None
        self._detect_language()
        if self._ready:
            self._process()

    def _open_dictionary(self) -> None:
        """
        Launch dictionary.
        """
        selected_word = self._check_dictionary_selected()
        if self._dictionary_window:
            if selected_word != '':
                self._dictionary_window.insert_word(selected_word)
            self._dictionary_window.root.lift()
            self._dictionary_window.set_lang(self._detected_lang_tag)
            return
        self._dictionary_window = gui_ut.DictionaryGUI((340, 300), self._cfg, self._detected_lang_tag, self)
        self._dictionary_window.insert_word(selected_word)
        self._dictionary_window.on_destroy = self._close_dictionary
        try:
            self._dictionary_window.root.update()
        except AttributeError:
            pass

    def _close_dictionary(self) -> None:
        """
        Close dictionary.
        """
        self._dictionary_window = None

    def _close(self) -> None:
        """
        Close the window.
        """
        if self._settings_window:
            self._settings_window.close()
        if self._dictionary_window:
            self._dictionary_window.close()
        self._root.destroy()

    def _about(self) -> None:
        """
        Show about window.
        """
        _, _, ver = self._check_version()
        # f'{self._cfg.lang("about_author")}: {pydetex.__author__}\n\n' \
        n_app_opened = self._cfg.get(self._cfg.CFG_TOTAL_OPENED_APP)
        n_word_processed = self._cfg.get(self._cfg.CFG_TOTAL_PROCESSED_WORDS)
        msg = f'PyDetex v{pydetex.version.ver}\n' \
              f'{ver}\n\n' \
              f'{self._cfg.lang("about_opened")}: {ut.format_number_d(n_app_opened, self._cfg.lang("format_d"))}\n' \
              f'{self._cfg.lang("about_processed")}: {ut.format_number_d(n_word_processed, self._cfg.lang("format_d"))}\n\n' \
              f'{pydetex.__copyright__}'

        messagebox.showinfo(title=self._cfg.lang('about'), message=msg)

    def _check_version(self) -> Tuple[bool, str, str]:
        """
        Check software version.

        :return: (Needs software update, new version, version text about)
        """
        is_outdated = False
        latest_version = '0.0.0'
        # noinspection PyBroadException
        try:
            is_outdated, latest_version = check_outdated('pydetex', str(pydetex.version.ver))
            if is_outdated:
                about_ver = self._cfg.lang('about_ver_upgrade').format(latest_version)
            else:
                about_ver = self._cfg.lang('about_ver_latest')
        except ValueError:
            about_ver = self._cfg.lang('about_ver_dev')
        except requests.exceptions.ConnectionError:
            about_ver = self._cfg.lang('about_ver_err_conn')
        except Exception:
            about_ver = self._cfg.lang('about_ver_err_unkn')
        return is_outdated, latest_version, about_ver

    def _check_version_event(self) -> None:
        """
        Check the version. If outdated, raise a popup.
        """
        outdated, latest, _ = self._check_version()
        if outdated:
            messagebox.showinfo(
                title=self._cfg.lang('version_upgrade_title'),
                message=self._cfg.lang('version_upgrade').format(latest)
            )


if __name__ == '__main__':
    PyDetexGUI().start()



"""
PyDetex
https://github.com/ppizarror/PyDetex

GUI UTILS
Provides utils for the gui.
"""

__all__ = [
    'BorderedFrame',
    'center_window',
    'CustomEntry',
    'DictionaryGUI',
    'make_label',
    'RichText',
    'SettingsWindow'
]

import tkinter as tk
from tkinter import ttk
from tkinter import font as tkfont
from tkinter import messagebox

from PyMultiDictionary import MultiDictionary
from typing import Callable, Tuple, Optional, Dict, Union, List, TYPE_CHECKING

if TYPE_CHECKING:
    from pydetex.gui import PyDetexGUI

import pydetex._fonts as fonts
import pydetex.utils as ut
from pydetex._gui_settings import Settings as _Settings


class SettingsWindow(object):
    """
    Settings window.
    """

    _cfg: '_Settings'
    _dict_langs: Dict[str, str]
    _dict_pipelines: Dict[str, str]
    _var_check_repetition: 'tk.BooleanVar'
    _var_check_repetition_stemming: 'tk.BooleanVar'
    _var_check_repetition_stopwords: 'tk.BooleanVar'
    _var_font_size: 'tk.StringVar'
    _var_lang: 'tk.StringVar'
    _var_output_font_format: 'tk.BooleanVar'
    _var_pipeline: 'tk.StringVar'
    _var_repetition_distance: 'tk.Entry'
    _var_repetition_ignore_words: 'tk.Text'
    _var_repetition_min_char: 'tk.Entry'
    _var_window_size: 'tk.StringVar'
    on_destroy: Optional[Callable[[], None]]
    root: 'tk.Tk'

    # noinspection PyProtectedMember
    def __init__(self, window_size: Tuple[int, int], cfg: '_Settings') -> None:
        """
        Constructor.

        :param window_size: Window size (width, height)
        :param cfg: Settings
        """
        self.root = tk.Tk()
        self.on_destroy = None
        self._cfg = cfg  # Store setting reference

        # Configure window
        self.root.title('Settings')
        self.root.minsize(width=window_size[0], height=window_size[1])
        self.root.resizable(width=False, height=False)
        center_window(self.root, window_size)
        self.root.protocol('WM_DELETE_WINDOW', self.close)
        if not ut.IS_OSX:
            self.root.iconbitmap(ut.RESOURCES_PATH + 'cog.ico')

        # Registers
        reg_int = self.root.register(ut.validate_int)

        # Main frame
        f0 = tk.Frame(self.root, border=5)
        f0.pack(fill='both')

        label_w = 17

        # Set languages
        f = tk.Frame(f0, border=0)
        f.pack(fill='both', pady=5)
        tk.Label(f, text=self._cfg.lang('cfg_lang'), width=label_w, anchor='w').pack(side=tk.LEFT, padx=(5, 9))

        self._dict_langs = {}
        for k in self._cfg._lang.get_available():
            self._dict_langs[self._cfg._lang.get(k, 'lang')] = k

        self._var_lang = tk.StringVar(self.root)
        self._var_lang.set(self._cfg._lang.get(cfg.get(cfg.CFG_LANG), 'lang'))  # default value

        pipe = tk.OptionMenu(f, self._var_lang, *list(self._dict_langs.keys()))
        pipe.focus()
        pipe.pack(side=tk.LEFT)

        # Window size
        f = tk.Frame(f0, border=0)
        f.pack(fill='both', pady=5)
        tk.Label(f, text=self._cfg.lang('cfg_window_size'), width=label_w, anchor='w').pack(side=tk.LEFT, padx=(5, 9))

        self._dict_window_sizes = {}
        for k in self._cfg._valid_window_sizes:
            self._dict_window_sizes[self._cfg.lang(k)] = k

        self._var_window_size = tk.StringVar(self.root)
        self._var_window_size.set(self._cfg.lang(cfg.get(cfg.CFG_WINDOW_SIZE, update=False)))  # default value

        windowsize = tk.OptionMenu(f, self._var_window_size, *list(self._dict_window_sizes.keys()))
        windowsize.pack(side=tk.LEFT)

        # Set pipelines
        f = tk.Frame(f0, border=0)
        f.pack(fill='both', pady=5)
        tk.Label(f, text=self._cfg.lang('cfg_pipeline'), width=label_w,
                 anchor='w').pack(side=tk.LEFT, padx=(5, 9 if ut.IS_OSX else 7))

        self._dict_pipelines = {}
        for k in self._cfg._available_pipelines:
            self._dict_pipelines[self._cfg.lang(k)] = k

        self._var_pipeline = tk.StringVar(self.root)
        self._var_pipeline.set(self._cfg.lang(cfg.get(cfg.CFG_PIPELINE, update=False)))  # default value

        pipe = tk.OptionMenu(f, self._var_pipeline, *list(self._dict_pipelines.keys()))
        pipe.pack(side=tk.LEFT)

        # Check repetition
        f_repetition = tk.LabelFrame(f0, text=self._cfg.lang('cfg_words_repetition'), bd=1, relief=tk.GROOVE)
        f_repetition.pack(fill='both')

        f = tk.Frame(f_repetition, border=0)
        f.pack(fill='both')
        tk.Label(f, text=self._cfg.lang('cfg_check'), width=label_w, anchor='w').pack(
            side=tk.LEFT,
            padx=(5 if ut.IS_OSX else 4, 4))
        self._var_check_repetition = tk.BooleanVar(self.root)
        self._var_check_repetition.set(cfg.get(cfg.CFG_CHECK_REPETITION))
        tk.Checkbutton(f, variable=self._var_check_repetition).pack(side=tk.LEFT)

        # Repetition min chars
        f = tk.Frame(f_repetition, border=0)
        f.pack(fill='both')
        tk.Label(f, text=self._cfg.lang('cfg_words_repetition_minchars'), width=label_w, anchor='w').pack(
            side=tk.LEFT,
            padx=(5 if ut.IS_OSX else 4, 5 if ut.IS_OSX else 9)
        )
        self._var_repetition_min_char = CustomEntry(f, self._cfg, validate='all', validatecommand=(reg_int, '%P'),
                                                    width=5)
        self._var_repetition_min_char.pack(side=tk.LEFT)
        self.root.after(100, lambda: self._var_repetition_min_char.insert(0, cfg.get(cfg.CFG_REPETITION_MIN_CHAR)))

        # Repetition distance
        f = tk.Frame(f_repetition, border=0)
        f.pack(fill='both')
        tk.Label(f, text=self._cfg.lang('cfg_words_repetition_distance'), width=label_w, anchor='w').pack(
            side=tk.LEFT,
            padx=(5 if ut.IS_OSX else 4, 5 if ut.IS_OSX else 9)
        )
        self._var_repetition_distance = CustomEntry(f, self._cfg, validate='all',
                                                    validatecommand=(reg_int, '%P'), width=5)
        self._var_repetition_distance.pack(side=tk.LEFT)
        self.root.after(100, self._var_repetition_distance.insert(0, cfg.get(cfg.CFG_REPETITION_DISTANCE)))

        # Repetition use stemming
        f = tk.Frame(f_repetition, border=0)
        f.pack(fill='both')
        tk.Label(f, text=self._cfg.lang('cfg_words_repetition_stemming'), width=label_w, anchor='w').pack(
            side=tk.LEFT,
            padx=(5 if ut.IS_OSX else 4, 4)
        )
        self._var_check_repetition_stemming = tk.BooleanVar(self.root)
        self._var_check_repetition_stemming.set(cfg.get(cfg.CFG_REPETITION_USE_STEMMING))
        tk.Checkbutton(f, variable=self._var_check_repetition_stemming).pack(side=tk.LEFT)

        # Repetition use stopwords
        f = tk.Frame(f_repetition, border=0)
        f.pack(fill='both')
        tk.Label(f, text=self._cfg.lang('cfg_words_repetition_stopwords'), width=label_w, anchor='w').pack(
            side=tk.LEFT,
            padx=(5 if ut.IS_OSX else 4, 4)
        )
        self._var_check_repetition_stopwords = tk.BooleanVar(self.root)
        self._var_check_repetition_stopwords.set(cfg.get(cfg.CFG_REPETITION_USE_STOPWORDS))
        tk.Checkbutton(f, variable=self._var_check_repetition_stopwords).pack(side=tk.LEFT)

        # Repetition ignore words
        f = tk.Frame(f_repetition, border=0)
        f.pack(fill='both')
        tk.Label(f, text=self._cfg.lang('cfg_words_repetition_ignorew'), width=label_w, anchor='w').pack(
            side=tk.LEFT,
            padx=(5 if ut.IS_OSX else 4, 5 if ut.IS_OSX else 9)
        )
        self._var_repetition_ignore_words = RichText(cfg, f, wrap='word', height=4,
                                                     highlightthickness=3 if ut.IS_OSX else 0,
                                                     highlightcolor='#426392', editable=True)
        self._var_repetition_ignore_words.pack(side=tk.LEFT, padx=(0, 5))
        self._var_repetition_ignore_words.insert(0.0, cfg.get(cfg.CFG_REPETITION_IGNORE_WORDS).strip())

        # End repetition
        f = tk.Frame(f_repetition, border=0, height=3 if ut.IS_OSX else 5)
        f.pack()
        f.pack_propagate(0)

        # Font format
        f = tk.Frame(f0, border=0, relief=tk.GROOVE)
        f.pack(fill='both')
        tk.Label(f, text=self._cfg.lang('cfg_font_format'), width=label_w,
                 anchor='w').pack(side=tk.LEFT, padx=(5, 9 if ut.IS_OSX else 7))
        self._var_output_font_format = tk.BooleanVar(self.root)
        self._var_output_font_format.set(cfg.get(cfg.CFG_OUTPUT_FONT_FORMAT))
        tk.Checkbutton(f, variable=self._var_output_font_format).pack(side=tk.LEFT)

        # Set font size
        f = tk.Frame(f0, border=0)
        f.pack(fill='both', pady=5)
        tk.Label(f, text=self._cfg.lang('cfg_font_size'), width=label_w,
                 anchor='w').pack(side=tk.LEFT, padx=(5, 9 if ut.IS_OSX else 7))

        self._var_font_size = tk.StringVar(self.root)
        self._var_font_size.set(cfg.get(cfg.CFG_FONT_SIZE))

        fontsize = tk.OptionMenu(f, self._var_font_size, *cfg._valid_font_sizes)
        fontsize.pack(side=tk.LEFT)

        # Save
        fbuttons = tk.Frame(f0)
        fbuttons.pack(side=tk.BOTTOM, expand=True)
        ut.Button(fbuttons, text=ut.button_text(self._cfg.lang('cfg_save')), command=self._save,
                  relief=tk.GROOVE).pack(pady=(12 if ut.IS_OSX else 8, 0))

        # Update
        self.root.update()

    def close(self) -> None:
        """
        Close the window.
        """
        if self.on_destroy:
            self.on_destroy()
        self.root.destroy()

    def _save(self) -> None:
        """
        Save the settings.
        """
        lang_value, current_lang = (self._dict_langs[self._var_lang.get()],
                                    self._cfg.get(self._cfg.CFG_LANG))
        windowsz_value, current_windowsz = (self._dict_window_sizes[self._var_window_size.get()],
                                            self._cfg.get(self._cfg.CFG_WINDOW_SIZE, update=False))
        fontsize_value, current_fontsize = (self._var_font_size.get(),
                                            self._cfg.get(self._cfg.CFG_FONT_SIZE))

        store: Tuple[Tuple[str, str, str], ...] = (
            (self._cfg.CFG_LANG, lang_value,
             self._cfg.lang('cfg_error_lang')),
            (self._cfg.CFG_WINDOW_SIZE, windowsz_value,
             self._cfg.lang('cfg_error_window_size')),
            (self._cfg.CFG_PIPELINE, self._dict_pipelines[self._var_pipeline.get()],
             self._cfg.lang('cfg_error_pipeline')),
            (self._cfg.CFG_CHECK_REPETITION, self._var_check_repetition.get(),
             self._cfg.lang('cfg_error_repetition')),
            (self._cfg.CFG_REPETITION_DISTANCE, self._var_repetition_distance.get(),
             self._cfg.lang('cfg_error_repetition_distance')),
            (self._cfg.CFG_REPETITION_IGNORE_WORDS, self._var_repetition_ignore_words.get(0.0, tk.END),
             self._cfg.lang('cfg_error_repetition_words')),
            (self._cfg.CFG_REPETITION_MIN_CHAR, self._var_repetition_min_char.get(),
             self._cfg.lang('cfg_error_repetition_chars')),
            (self._cfg.CFG_REPETITION_USE_STEMMING, self._var_check_repetition_stemming.get(),
             self._cfg.lang('cfg_error_stemming')),
            (self._cfg.CFG_REPETITION_USE_STOPWORDS, self._var_check_repetition_stopwords.get(),
             self._cfg.lang('cfg_error_stopwords')),
            (self._cfg.CFG_OUTPUT_FONT_FORMAT, self._var_output_font_format.get(),
             self._cfg.lang('cfg_error_output_format')),
            (self._cfg.CFG_FONT_SIZE, fontsize_value,
             self._cfg.lang('cfg_error_font_size'))
        )

        # Set values
        do_close = True
        for cfg in store:
            try:
                self._cfg.set(cfg[0], cfg[1])
            except ValueError:
                messagebox.showerror('Error', cfg[2])
                do_close = False

        # Check if lang has changed
        if lang_value != current_lang or int(fontsize_value) != current_fontsize or \
                windowsz_value != current_windowsz:
            messagebox.showinfo(title=self._cfg.lang('reload_message_title'),
                                message=self._cfg.lang('reload_message_message'))

        # Save
        self._cfg.save()
        if do_close:
            self.close()


# noinspection PyProtectedMember
class DictionaryGUI(object):
    """
    Dictionary gui.
    """

    _ant: 'tk.Button'
    _cfg: '_Settings'
    _dictionary: 'MultiDictionary'
    _lang: str
    _main: 'PyDetexGUI'
    _mean: 'tk.Button'
    _syn: 'tk.Button'
    root: 'tk.Tk'

    # noinspection PyProtectedMember
    def __init__(self, window_size: Tuple[int, int], cfg: '_Settings', lang: str, main: 'PyDetexGUI') -> None:
        """
        Constructor.

        :param window_size: Window size (width, height)
        :param cfg: Settings
        :param lang: Dictionary lang
        :param main: Main object
        """
        self.root = tk.Tk()
        self.on_destroy = None

        self._cfg = cfg  # Store setting reference
        self._dictionary = MultiDictionary()
        self._main = main

        # Configure window
        self.root.minsize(width=window_size[0], height=window_size[1])
        self.root.resizable(width=False, height=False)
        center_window(self.root, window_size)
        self.root.protocol('WM_DELETE_WINDOW', self.close)
        if not ut.IS_OSX:
            self.root.iconbitmap(ut.RESOURCES_PATH + 'dictionary.ico')

        # Main frame
        f0 = tk.Frame(self.root, border=5)
        f0.pack(fill='both')

        # Word
        f = tk.Frame(f0, border=0)
        f.pack(fill='both', pady=5)
        tk.Label(f, text=self._cfg.lang('dictionary_word'), width=5, anchor='w').pack(side=tk.LEFT, padx=(5, 9))

        self._word = CustomEntry(f, self._cfg, width=30 if ut.IS_OSX else 50)
        self._word.pack(side=tk.LEFT)

        # Add commands
        f = tk.Frame(f0, border=0)
        f.pack(fill='both', pady=(5, 2))
        btn_width = 9
        btn_pad = 5 if ut.IS_OSX else 15
        self._mean = tk.Button(f, text=self._cfg.lang('dictionary_meaning'), command=self._meaning, relief=tk.GROOVE,
                               width=btn_width)
        self._mean.pack(padx=(0, btn_pad), side=tk.LEFT)
        self._syn = tk.Button(f, text=self._cfg.lang('dictionary_synonym'), command=self._synonym, relief=tk.GROOVE,
                              width=btn_width)
        self._syn.pack(padx=(0, btn_pad), side=tk.LEFT)
        self._ant = tk.Button(f, text=self._cfg.lang('dictionary_antonym'), command=self._antonym, relief=tk.GROOVE,
                              width=btn_width)
        self._ant.pack(padx=0, side=tk.LEFT)
        # f = tk.Frame(f0, border=0)
        # f.pack(fill='both', pady=(0, 5))
        # tk.Button(f, text=self._cfg.lang('dictionary_translation'), command=self._meaning, relief=tk.GROOVE, width=7).pack()

        # Out text
        f = tk.Frame(f0, border=0, width=window_size[0], height=50)
        f.pack(fill='both', pady=(5, 0))
        f.pack_propagate()

        hthick, hcolor = 3 if ut.IS_OSX else 1, '#426392' if ut.IS_OSX else '#475aff'

        self._text_out = RichText(self._cfg, f, wrap='word', highlightthickness=hthick,
                                  highlightcolor=hcolor, copy=True)
        self._text_out.bind('<Key>', self._process_out_key)
        self._text_out.pack(fill='both')
        self._text_out['state'] = tk.DISABLED

        # Set the language
        self.set_lang(lang)

    def set_lang(self, lang: str) -> None:
        """
        Set language.

        :param lang: Language
        """
        if lang not in self._dictionary._langs.keys():
            self.close()
        lang_name = self._dictionary.get_language_name(lang, self._cfg.get(self._cfg.CFG_LANG))
        self.root.title(f'{self._cfg.lang("dictionary")} [{lang_name}]')
        self._lang = lang

        # Check language capability
        syn, mean, _, ant = self._dictionary._langs[lang]
        self._syn['state'] = tk.NORMAL if syn else tk.DISABLED
        self._mean['state'] = tk.NORMAL if mean else tk.DISABLED
        self._ant['state'] = tk.NORMAL if ant else tk.DISABLED

    @staticmethod
    def _process_out_key(event: 'tk.Event') -> Optional['tk.Event']:
        """
        Process out keys.

        :param event: Event
        :return: Event
        """
        # noinspection PyUnresolvedReferences
        if event.char == '':
            return event

    def _get_word(self) -> str:
        """
        Get the word.

        :return: Word
        """
        return self._word.get().strip()

    def _meaning(self) -> None:
        """
        Word meaning.
        """
        self._main._status(self._cfg.lang('dictionary_querying'))
        style, mean, wiki = self._dictionary.meaning(self._lang, self._get_word())
        self._main._status_clear()
        if mean == '':
            return self._write(self._cfg.lang('dictionary_no_results'))
        style = ', '.join(style)
        self._text_out['state'] = tk.NORMAL
        self._text_out.clear()
        self._text_out.insert('end', f'{style}\n\n', 'bold')
        self._text_out.insert('end', f'{self._cfg.lang("dictionary_meaning")}:\n', 'italic')
        self._text_out.insert('end', mean, 'normal')
        if wiki != '':
            self._text_out.insert('end', '\n\n', 'normal')
        self._text_out.insert('end', f'{self._cfg.lang("dictionary_wikipedia")}:\n', 'italic')
        self._text_out.insert('end', f'{wiki}', 'normal')
        self._text_out['state'] = tk.DISABLED

    def _synonym(self) -> None:
        """
        Word synonym.
        """
        self._main._status(self._cfg.lang('dictionary_querying'))
        syn = self._dictionary.synonym(self._lang, self._get_word())
        self._main._status_clear()
        if len(syn) == 0:
            return self._write(self._cfg.lang('dictionary_no_results'))
        self._write(', '.join(syn))

    def _antonym(self) -> None:
        """
        Word synonym.
        """
        self._main._status(self._cfg.lang('dictionary_querying'))
        syn = self._dictionary.antonym(self._lang, self._get_word())
        self._main._status_clear()
        if len(syn) == 0:
            return self._write(self._cfg.lang('dictionary_no_results'))
        self._write(', '.join(syn))

    def _write(self, text: str) -> None:
        """
        Write a text.

        :param text: Text
        """
        self._text_out['state'] = tk.NORMAL
        self._text_out.delete(0.0, tk.END)
        self._text_out.insert(0.0, text)
        self._text_out['state'] = tk.DISABLED

    def insert_word(self, word: str) -> None:
        """
        Insert a word to the fields.

        :param word: Word
        """
        self._word.delete(0, tk.END)
        stemmer = ut.make_stemmer(self._lang)
        if stemmer is not None:
            word = stemmer.stem(word)
        self.root.after(50, lambda: self._word.insert(0, word))

    def close(self) -> None:
        """
        Close the window.
        """
        if self.on_destroy:
            self.on_destroy()
        self.root.destroy()


# noinspection PyShadowingNames,PyMissingOrEmptyDocstring
class RichText(tk.Text):
    """
    Rich text.
    """

    _default_font: 'tkfont.Font'
    _default_size: int
    _em: int

    def __init__(self, cfg: '_Settings', *args, **kwargs):
        font_size = kwargs.pop('font_size', 11)
        editable = kwargs.pop('editable', False)
        copy = kwargs.pop('copy', False)
        if editable:
            kwargs['undo'] = True

        super().__init__(*args, **kwargs)
        self._default_font = tkfont.nametofont(self.cget('font'))
        self._default_font.configure(size=font_size)

        self._em = self._default_font.measure('m')
        self._default_size = self._default_font.cget('size')

        # Editable gui
        if editable:
            EditableTextGUI(self, cfg)
        elif copy:
            CopyTextGUI(self, cfg)

        # Configure specials
        self.tag_configure('bullet', lmargin1=self._em,
                           lmargin2=self._em + self._default_font.measure('\u2022 '))

        # Add fonts
        for tag in fonts.FONT_PROPERTIES.keys():
            style = fonts.FONT_PROPERTIES[tag]
            if style is None:
                continue
            self._add_font(tag, **style)

    def _add_font(self, tag: str, **kwargs) -> None:
        font = tkfont.Font(**self._default_font.configure())

        # Update kwargs for font
        if 'size' not in kwargs.keys():
            kwargs['size'] = self._default_size
        else:
            kwargs['size'] = int(kwargs['size'] * self._default_size)

        # Move kwargs to tag
        if kwargs.pop('ignore_font', False):
            tag_kwargs = {}
        else:
            tag_kwargs = {'font': font}
        for t in ['foreground', 'background', 'spacing3']:
            if t in kwargs.keys():
                tag_kwargs[t] = kwargs.pop(t)
        if 'spacing3' in tag_kwargs.keys():
            tag_kwargs['spacing3'] *= self._default_size

        font.configure(**kwargs)
        self.tag_configure(tag, **tag_kwargs)

    def insert_bullet(self, index: float, text: str) -> None:
        """
        Inserts a bullet.

        :param index: Position
        :param text: Text
        """
        self.insert(index, f'\u2022 {text}', 'bullet')

    def insert_highlighted_text(self, s: str, clear: bool = False, font_format: bool = True) -> None:
        """
        Insert a highlighted text.

        :param s: Text
        :param clear: Clear before insert
        :param font_format: If False, use normal font instead
        """
        # Write results and split tags
        if clear:
            self.clear()
        for t in ut.split_tags(s, list(fonts.FONT_TAGS.values())):
            tag, text = t
            self.insert('end', text, fonts.TAGS_FONT[tag] if font_format else 'normal')

    def clear(self) -> None:
        """
        Clears the text.
        """
        self.delete(0.0, tk.END)


# noinspection PyUnresolvedReferences,PyUnusedLocal
class EditableTextGUI(object):
    """
    Editable gui text.
    """

    _w: Union['tk.Entry', 'tk.Text', 'RichText']

    def __init__(self, widget: Union['tk.Entry', 'tk.Text', 'RichText'], cfg: '_Settings') -> None:
        """
        Constructor.

        :param widget: Widget object
        :param cfg: Settings
        """
        self._w = widget

        self.changes = ['']
        self.steps = int()

        self.context_menu = tk.Menu(self._w, tearoff=0)
        self.context_menu.add_command(label=cfg.lang('menu_cut'))
        self.context_menu.add_command(label=cfg.lang('menu_copy'))
        self.context_menu.add_command(label=cfg.lang('menu_paste'))

        self._w.bind('<Button-3>' if not ut.IS_OSX else '<Button-2>', self.popup)
        self._w.bind('<Control-z>', self.undo)
        self._w.bind('<Control-y>', self.redo)
        self._w.bind('<Key>', self.add_changes)

        if ut.IS_OSX:
            self._w.bind('<Control-c>', lambda _: self._w.event_generate('<<Copy>>'))
            self._w.bind('<Control-v>', lambda _: self._w.event_generate('<<Paste>>'))

    def popup(self, event: 'tk.Event') -> None:
        """
        Raise popup.

        :param event: Event
        """
        self.context_menu.post(event.x_root, event.y_root)
        self.context_menu.entryconfigure('Cut', command=lambda: self._w.event_generate('<<Cut>>'))
        self.context_menu.entryconfigure('Copy', command=lambda: self._w.event_generate('<<Copy>>'))
        self.context_menu.entryconfigure('Paste', command=lambda: self._w.event_generate('<<Paste>>'))

    def undo(self, event: Optional['tk.Event'] = None) -> None:
        """
        Undo operation.

        :param event: Event
        """
        if self.steps != 0:
            self.steps -= 1
            self._w.delete(0, tk.END)
            self._w.insert(tk.END, self.changes[self.steps])

    def redo(self, event: Optional['tk.Event'] = None) -> None:
        """
        Redo operation.

        :param event: Event
        """
        if self.steps < len(self.changes):
            self._w.delete(0, tk.END)
            self._w.insert(tk.END, self.changes[self.steps])
            self.steps += 1

    def add_changes(self, event: Optional['tk.Event'] = None) -> None:
        """
        Add changes.

        :param event: Event
        """
        if self._w.get() != self.changes[-1]:
            self.changes.append(self._w.get())
            self.steps += 1


# noinspection PyUnresolvedReferences,PyUnusedLocal
class CopyTextGUI(object):
    """
    Enables a menu for only copy.
    """

    _w: Union['tk.Entry', 'tk.Text', 'RichText']

    def __init__(self, widget: Union['tk.Entry', 'tk.Text', 'RichText'], cfg: '_Settings') -> None:
        """
        Constructor.

        :param widget: Widget object
        :param cfg: Settings
        """
        self._w = widget

        self.context_menu = tk.Menu(self._w, tearoff=0)
        self.context_menu.add_command(label=cfg.lang('menu_copy'))

        self._w.bind('<Button-3>' if not ut.IS_OSX else '<Button-2>', self.popup)

        if ut.IS_OSX:
            self._w.bind('<Control-c>', lambda _: self._w.event_generate('<<Copy>>'))

    def popup(self, event: 'tk.Event') -> None:
        """
        Raise popup.

        :param event: Event
        """
        self.context_menu.post(event.x_root, event.y_root)
        self.context_menu.entryconfigure('Copy', command=lambda: self._w.event_generate('<<Copy>>'))


class CustomEntry(tk.Entry):
    """
    Entry with undo/redo and menu.
    """

    def __init__(self, parent, cfg: '_Settings', *args, **kwargs):
        tk.Entry.__init__(self, parent, *args, **kwargs)
        EditableTextGUI(self, cfg)


class BorderedFrame(tk.Frame):
    """
    Bordered frame widget.
    """

    def __init__(self, master, bordercolor=None, borderleft=0, bordertop=0, borderright=0, borderbottom=0,
                 interiorwidget=tk.Frame, **kwargs):
        tk.Frame.__init__(self, master, background=bordercolor, bd=0, highlightthickness=0)

        self.interior = interiorwidget(self, **kwargs)
        self.interior.pack(padx=(borderleft, borderright), pady=(bordertop, borderbottom))


def center_window(root: 'tk.Tk', window_size: Union[Tuple[int, int], List[int]]) -> None:
    """
    Center window.

    :param root: Window object
    :param window_size: Window size
    """
    if ut.IS_OSX:
        root.geometry('%dx%d+%d+%d' % (window_size[0], window_size[1],
                                       (root.winfo_screenwidth() - window_size[0]) / 2,
                                       (root.winfo_screenheight() - window_size[1]) / 2 - 25))
    else:
        root.geometry('%dx%d+%d+%d' % (window_size[0], window_size[1],
                                       (root.winfo_screenwidth() - window_size[0]) / 2,
                                       (root.winfo_screenheight() - window_size[1]) / 2))


def make_label(master, h, w, side, *args, pad=(0, 0, 0, 0), separator=False, **kwargs) -> 'tk.Label':
    """
    Makes a label with defined width/height.

    :param master: Master object
    :param h: Height in pixels
    :param w: Width in pixels
    :param side: Packing side
    :param args: Label arguments
    :param pad: Padding (top, right, bottom, left)
    :param separator: Add separator
    :param kwargs: Optional keyword-arguments
    :return: Label
    """
    f = tk.Frame(master, height=int(h), width=int(w))
    f.pack_propagate(0)  # don't shrink
    f.pack(side=side)
    label = tk.Label(f, *args, **kwargs)
    if w > 0:
        label.pack(fill=tk.BOTH, expand=1, padx=(int(pad[3]), int(pad[1])), pady=(int(pad[0]), int(pad[2])))
        if separator:
            ttk.Separator(master, orient='vertical').pack(side=tk.LEFT, fill='y')
    return label



"""
PyDetex
https://github.com/ppizarror/pydetex

PARSERS
Defines parsers, which perform a single task for removal LaTex things.
"""

__all__ = [
    'find_str',
    'FONT_FORMAT_SETTINGS',
    'process_cite',
    'process_inputs',
    'process_labels',
    'process_quotes',
    'process_ref',
    'remove_commands_char',
    'remove_commands_param',
    'remove_commands_param_noargv',
    'remove_comments',
    'remove_common_tags',
    'remove_tag',
    'simple_replace'
]

import os
from typing import List, Tuple, Union

# Files
_NOT_FOUND_FILES = []
_PRINT_LOCATION = False
_TAG_FILE_ERROR = '|FILEERROR|'

# Parser font format
FONT_FORMAT_SETTINGS = {
    'cite': '',
    'normal': '',
    'ref': ''
}

import pydetex.utils as ut


def _find_str(s: str, char: str) -> int:
    """
    Finds a sequence within a string, and returns the position. If not exists, returns -1.

    :param s: String
    :param char: Sequence
    :return: Position
    """
    index = 0

    if char in s:
        c = char[0]
        for ch in s:
            if ch == c:
                if s[index:index + len(char)] == char:
                    return index

            index += 1

    return -1


def find_str(s: str, char: Union[str, List[str], Tuple[str, ...]]) -> int:
    """
    Finds a sequence within a string, and returns the position. If not exists, returns -1.

    :param s: String
    :param char: Sequence or List of sequences
    :return: Position
    """
    if isinstance(char, str):
        return _find_str(s, char)
    else:
        for ch in char:
            j = _find_str(s, ch)
            if j != -1:
                return j
    return -1


def remove_tag(s: str, tagname: str) -> str:
    """
    Removes a latex tag code.

    :param s: String
    :param tagname: Tag code
    :return: String without tags
    """
    tagname = '\\' + tagname
    tagadd = 1
    if '{' not in tagname:
        tagname += '{'
        tagadd = 0
    while True:
        k = find_str(s, tagname)
        if k == -1:  # No more tags, return
            return s
        deep = 0
        f = False
        for j in range(len(s)):
            if s[k + j] == '{':
                deep += 1
                f = True
                continue
            if s[k + j] == '}':
                deep -= 1
            if deep == 0 and f:
                # update s
                s = s[:k] + s[k + len(tagname) + tagadd:k + j] + s[k + j + 1:]
                break


def remove_common_tags(s: str) -> str:
    """
    Remove common tags from string.

    :param s: Text
    :return: Text without tags
    """
    for tag in [
        'bigp',
        'chapter',
        'emph',
        'section',
        'subsection',
        'subsubsection',
        'subsubsubsection',
        'textbf',
        'textit',
        'texttt',

        # Enviroments
        'begin{itemize}',
        'end{itemize}'
    ]:
        s = remove_tag(s, tag)
    return s


def process_cite(s: str) -> str:
    """
    Transforms all cites to a text-based with numbers. For example,
    'This is from \\cite{Pizarro}' to 'This is from [1].

    :param s: String
    :return:
    """
    cites = {}
    look = ['\\cite*{', '\\citet*{', '\\citep*{', '\\cite{', '\\citet{', '\\citep{']
    k = -1
    while True:
        for j in look.copy():
            k = find_str(s, j)
            if k == -1:
                look.remove(j)
            else:
                break
        if k == -1:
            return s
        for j in range(len(s)):
            if s[k + j] == '}':
                c = s[k + 6:k + j]
                for w in c.split(','):
                    if w not in cites.keys():
                        cites[w] = len(cites.keys()) + 1
                    c = c.replace(w, str(cites[w]))
                s = s[:k] + FONT_FORMAT_SETTINGS['cite'] + '[' + c + ']' + FONT_FORMAT_SETTINGS['normal'] + s[
                                                                                                            k + j + 1:]
                break


def process_labels(s: str) -> str:
    """
    Removes labels.

    :param s: String
    :return:
    """
    while True:
        k = find_str(s, '\\label{')
        if k == -1:
            return s
        for j in range(len(s)):
            if s[k + j] == '}':
                s = s[:k] + s[k + j + 1:]
                break


def process_ref(s) -> str:
    """
    Process references, same as cites, replaces by numbers.

    :param s: String
    :return: String with numbers instead of references.
    """
    r = 1
    while True:
        k = find_str(s, '\\ref{')
        if k == -1:
            return s
        for j in range(len(s)):
            if s[k + j] == '}':
                s = s[:k] + FONT_FORMAT_SETTINGS['ref'] + str(r) + FONT_FORMAT_SETTINGS['normal'] + s[k + j + 1:]
                r += 1
                break


def process_quotes(s) -> str:
    """
    Process quotes.

    :param s: String
    :return: String with "quotes"
    """
    while True:
        k = find_str(s, ['\\quotes{', '\\doublequotes{'])
        if k == -1:
            return s
        m = 0
        for j in range(len(s)):
            if s[k + j] == '{':
                m = j
            if s[k + j] == '}':
                s = s[:k] + '"' + s[k + m + 1:k + j] + '"' + s[k + j + 1:]
                break


def remove_comments(s: str) -> str:
    """
    Remove comments from text.

    :param s: Text
    :return: Text without comments
    """
    symbol = '|COMMENTPERCENTAGESYMBOL|'
    s = s.replace('  ', ' ')
    s = s.replace('\\%', symbol)
    k = s.split('\n')
    for r in range(len(k)):
        k[r] = k[r].strip()  # Strips all text
    line_merge = []
    for r in range(len(k)):
        sp = k[r].split('%')
        k[r] = sp[0]  # Removes all comments from list
        line_merge.append(len(sp) > 1)
    line_merge.append(False)
    k.append('')
    for r in range(len(k)):
        if line_merge[r] and not line_merge[r + 1] and k[r + 1] != '':
            line_merge[r + 1] = True
    new_k = []
    j = 0
    merged_str = ''
    while True:  # Merge comment lines
        if not line_merge[j]:
            if merged_str != '':  # Add current merged str
                new_k.append(merged_str)
                merged_str = ''
            new_k.append(k[j])
        else:
            merged_str += k[j]
        if j == len(k) - 1:
            break
        j += 1
    if merged_str != '':
        new_k.append(merged_str)
    k = new_k
    w = []  # Removes duplicates '' lines to single ''
    last = ''
    for j in k:
        if j == '' and j == last:
            pass
        else:
            w.append(j)
        last = j
    if len(w) > 0 and w[-1] == '':  # Removes last space
        w.pop()
    s = '\n'.join(w)
    s = s.replace(symbol, '%').strip()
    return s


def simple_replace(s: str) -> str:
    """
    Replace simple tokens.

    :param s: String
    :return: String with replaced items
    """
    library: List[Tuple[str, str]] = [
        # Common
        ('\\item', '-'),
        ('--', '–'),
        ('\\\\', '\n'),
        ('\\ ', ' '),

        # Letters
        ('ﬁ', 'fi'),

        # Greek
        ('\\alpha', 'α'),
        ('\\beta', 'β'),
        ('\\chi', 'χ'),
        ('\\delta', 'δ'),
        ('\\Delta', 'Δ'),
        ('\\epsilon', 'ϵ'),
        ('\\eta', 'η'),
        ('\\gamma', 'γ'),
        ('\\Gamma', 'Γ'),
        ('\\iota', 'ι'),
        ('\\kappa', 'κ'),
        ('\\lambda', 'λ'),
        ('\\Lambda', 'Λ'),
        ('\\mu', 'μ'),
        ('\\nu', 'ν'),
        ('\\omega', 'ω'),
        ('\\Omega', 'Ω'),
        ('\\phi', 'φ'),
        ('\\Phi', 'Φ'),
        ('\\pi', 'π'),
        ('\\Pi', 'Π'),
        ('\\psi', 'ψ'),
        ('\\Psi', 'Ψ'),
        ('\\rho', 'ρ'),
        ('\\sigma', 'σ'),
        ('\\Sigma', 'Σ'),
        ('\\tau', 'τ'),
        ('\\theta', 'θ'),
        ('\\Theta', 'Θ'),
        ('\\upsilon', 'υ'),
        ('\\varepsilon', 'ε'),
        ('\\varphi', 'φ'),
        ('\\varrho', 'ϱ'),
        ('\\vartheta', '𝜗'),
        ('\\xi', 'ξ'),
        ('\\Xi', 'Ξ'),
        ('\\zeta', 'ζ'),

        # Arrows
        ('\\leftarrow', '←'),
        ('\\rightarrow', '→'),
        ('\\Leftarrow', '⇐'),
        ('\\Rightarrow', '⇒'),
        ('\\uparrow', '↑'),
        ('\\downarrow', '↓'),
        ('\\Uparrow', '⇑'),
        ('\\Downarrow', '⇓'),
        ('\\leftrightarrow', '↔'),
        ('\\longleftarrow', '⟵'),
        ('\\longrightarrow', '⟶')
    ]
    for w in library:
        s = s.replace(w[0], w[1])
    return s


def _load_file(f: str, path: str) -> str:
    """
    Try to load a file.

    :param f: Filename
    :param path: Path to look from
    :return: File contents
    """
    try:
        fo = open(path + f, 'r')
        s = '\n'.join(fo.readlines())
        fo.close()
        return s
    except FileNotFoundError:
        return _TAG_FILE_ERROR


def _os_listfolder() -> List[str]:
    """
    Returns the folders from the current path.

    :return: Folder list
    """
    dirs = os.listdir('./')
    folders = []
    for k in dirs:
        if os.path.isdir(k):
            folders.append(k + '/')
    return folders


def process_inputs(s: str) -> str:
    """
    Process inputs, and try to copy the content.

    :param s: Text with inputs
    :return: Text copied with data
    """
    global _PRINT_LOCATION, _NOT_FOUND_FILES
    tx = ''
    while True:
        k = find_str(s, '\\input{')
        if k == -1:
            return s.replace('|INPUTFILETAG', '\\input{')
        m = 0
        for j in range(len(s)):
            if s[k + j] == '{':
                m = j
            if s[k + j] == '}':
                tex_file = s[k + m + 1:k + j]
                if '.tex' not in tex_file:
                    tex_file += '.tex'
                if tex_file not in _NOT_FOUND_FILES:
                    if not _PRINT_LOCATION:
                        print(f'Current path location: {os.getcwd()}')
                        _PRINT_LOCATION = True
                    print(f'Detected file {tex_file}:')

                    # Get folder locations
                    folders = _os_listfolder()
                    folders.insert(0, '../')
                    folders.insert(0, './')
                    for f in folders:
                        tx = _load_file(tex_file, f)
                        if tx == _TAG_FILE_ERROR:
                            print(f'\tFile not found in {f}')
                        else:
                            break
                    if tx == _TAG_FILE_ERROR:
                        _NOT_FOUND_FILES.append(tex_file)
                        s = s[:k] + '|INPUTFILETAG' + s[k + m + 1:]
                    else:
                        print('\tFile found and loaded')
                        s = s[:k] + tx + s[k + j + 1:]
                else:
                    s = s[:k] + '|INPUTFILETAG' + s[k + m + 1:]
                break


def remove_commands_char(s: str, chars: Union[Tuple[str, str], str], ignore_escape: bool = True) -> str:
    """
    Remove all char commands.

    :param s: String
    :param chars: Chars (start, end)
    :param ignore_escape: Ignores \char
    :return: Code with removed chars
    """
    tex_tags = ut.find_tex_command_char(s, chars, ignore_escape)
    if len(tex_tags) == 0:
        return s
    new_s = ''
    k = 0  # Moves through tags

    for i in range(len(s)):
        if k < len(tex_tags):
            if i < tex_tags[k][0]:
                new_s += s[i]
            elif i < tex_tags[k][1]:
                pass
            else:  # advance to other tag
                k += 1
        else:
            new_s += s[i]

    return new_s


def remove_commands_param(s: str) -> str:
    """
    Remove all commands with params.

    :param s: String
    :return: Code with removed chars
    """
    tex_tags = ut.find_tex_commands(s)
    if len(tex_tags) == 0:
        return s
    new_s = ''
    k = 0  # Moves through tags

    for i in range(len(s)):
        if k < len(tex_tags):
            if i < tex_tags[k][0]:
                new_s += s[i]
            elif i < tex_tags[k][3] + 1:
                pass
            else:  # advance to other tag
                k += 1
        else:
            new_s += s[i]

    return new_s


def remove_commands_param_noargv(s: str) -> str:
    """
    Remove all commands without arguments.

    :param s: String
    :return: Code with removed chars
    """
    tex_tags = ut.find_tex_commands_noargv(s)
    if len(tex_tags) == 0:
        return s
    new_s = ''
    k = 0  # Moves through tags

    for i in range(len(s)):
        if k < len(tex_tags):
            if i < tex_tags[k][0]:
                new_s += s[i]
            elif i < tex_tags[k][1]:
                pass
            else:  # advance to other tag
                k += 1
        else:
            new_s += s[i]

    return new_s
